/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WADJOINT_H
#define WADJOINT_H

#include "dart.h"
#include "wObject.h"
#include "wTimers.h"

#include <iostream>
#include <vector>
#include <memory>
#include <Eigen/Sparse>

namespace dart
{

/**
 * @brief Adjoint solver and API
 * @authors Adrien Crovato
 * @todo add interface for Eigen to avoid using Eigen::Map
 */
class DART_API Adjoint : public fwk::wSharedObject
{
public:
    std::shared_ptr<Newton> sol;                 ///< Newton solver
    std::shared_ptr<tbox::MshDeform> morph;      ///< mesh morpher
    std::shared_ptr<tbox::LinearSolver> alinsol; ///< linear solver (adjoint aero)
    std::shared_ptr<tbox::LinearSolver> mlinsol; ///< linear solver (adjoint mesh)

    int nthreads; ///< number of threads for the assembly
    int verbose;  ///< display more info

    std::vector<double> lamClFlo;        ///< lift flow adjoint
    std::vector<double> lamCdFlo;        ///< drag flow adjoint
    std::vector<Eigen::Vector3d> dClMsh; ///< lift gradient with respect to mesh
    std::vector<Eigen::Vector3d> dCdMsh; ///< drag gradient with respect to mesh
    double dClAoa;                       ///< lift gradient with respect to angle of attack
    double dCdAoa;                       ///< drag gradient with respect to angle of attack

private:
    Eigen::SparseMatrix<double, Eigen::RowMajor> dRx_X; ///< mesh Jacobian
    Eigen::SparseMatrix<double, Eigen::RowMajor> dRu_U; ///< flow Jacobian
    Eigen::VectorXd dRu_A;                              ///< partial gradients of flow residual wrt aoa
    Eigen::SparseMatrix<double, Eigen::RowMajor> dRu_X; ///< partial gradients of flow residual wrt mesh
    Eigen::SparseMatrix<double, Eigen::RowMajor> dL_U;  ///< partial gradients of loads wrt flow
    Eigen::SparseMatrix<double, Eigen::RowMajor> dL_X;  ///< partial gradients of loads wrt mesh
    fwk::Timers tms;                                    ///< internal timers

public:
    Adjoint(std::shared_ptr<Newton> _sol, std::shared_ptr<tbox::MshDeform> _morph);
    virtual ~Adjoint() { std::cout << "~Adjoint()\n"; }

    void run();
    void save(tbox::MshExport *mshWriter, int n = 0);

    // Partial gradients of the mesh residuals
    void linearizeMesh();
    std::vector<double> solveMesh(std::vector<double> const &dMesh);
    std::vector<double> computeJacVecMesh(std::vector<double> const &dR);
    // Partial gradients of the flow residuals
    void linearizeFlow();
    std::vector<double> solveFlow(std::vector<double> const &dFlow);
    std::vector<double> computeJacVecFlow(std::vector<double> const &dR);
    double computeJacVecFlowAoa(std::vector<double> const &dR);
    std::vector<double> computeJacVecFlowMesh(std::vector<double> const &dR);
    // Partial gradients of the loads
    void linearizeLoads(Body const &bnd);
    std::vector<double> computeJacVecLoadsFlow(std::vector<double> const &dL);
    std::vector<double> computeJacVecLoadsMesh(std::vector<double> const &dL);
    // Partial gradients of the coefficients
    std::vector<std::vector<double>> computeGradientCoefficientsFlow();
    std::vector<double> computeGradientCoefficientsAoa();
    std::vector<std::vector<double>> computeGradientCoefficientsMesh();

#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif

private:
    // Partial gradients wrt with respect to flow variables
    void buildGradientLoadsFlow(Eigen::SparseMatrix<double, Eigen::RowMajor> &dL, Body const &bnd);
    void buildGradientCoefficientsFlow(Eigen::RowVectorXd &dCl, Eigen::RowVectorXd &dCd, Body const &bnd);
    // Partial gradients with respect to angle of attack
    void buildGradientResidualAoa(Eigen::VectorXd &dR);
    void buildGradientLoadsAoa(double &dCl, double &dCd);
    // Partial gradients with respect to mesh coordinates
    void buildGradientResidualMesh(Eigen::SparseMatrix<double, Eigen::RowMajor> &dR);
    void buildGradientLoadsMesh(Eigen::SparseMatrix<double, Eigen::RowMajor> &dL, Body const &bnd);
    void buildGradientCoefficientsMesh(Eigen::RowVectorXd &dCl, Eigen::RowVectorXd &dCd, Body const &bnd);
};

} // namespace dart
#endif //WADJOINT_H
