#!/usr/bin/env python3 
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute lifting (linear or nonlinear) viscous flow around a NACA 0012
# Amaury Bilocq
#
# Test the viscous-inviscid interaction scheme
# Reference to the master's thesis: http://hdl.handle.net/2268/252195
# Reference test cases with Naca0012 (different from master's thesis):
# 1) Incompressible: Re = 1e7, M_inf = 0, alpha = 5°, msTE = 0.01, msLE = 0.01
#    -> nIt = 27, Cl = 0.55, Cd = 0.0058, xtrTop = 0.087, xtrBot = 0.741
#    -> msLe = 0.001, xtrTop = 0.061
# 2) Compressible: Re = 1e7, M_inf = 5, alpha = 5°, msTE = 0.01, msLE = 0.01
#    -> nIt = 33, Cl = 0.65, Cd = 0.0063, xtrTop = 0.057, xtrBot = 0.740
# 3) Separated: Re = 1e7, M_inf = 0, alpha = 12°, msTE = 0.01, msLE = 0.001
#    -> nIt = 43, Cl = 1.30 , Cd = 0.0108, xtrTop = 0.010, xtrBot = 1
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import dart.utils as floU
import dart.default as floD

import dart.pyVII.vCoupler as floVC
import dart.pyVII.vUtils as viscU

import tbox.utils as tboxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

import math

def Case(cs = 0):

    if cs == 0:
        Re = 1e7
        alpha = 5.*math.pi/180
        M = 0.
        arf = 'models/n0012.geo'
    elif cs == 1:
        Re = 1e7
        alpha = 12.*math.pi/180
        M = 0.
        arf = 'models/n0012.geo'
    elif cs == 2:
        Re = 1e7
        alpha = 2.*math.pi/180
        M = .715
        arf = 'models/rae2822.geo'
    else:
        Re = 1e7
        alpha = 0.
        M = 0.
        arf = 'models/n0012.geo'

    return Re, alpha, M, arf



def main():
    print('Start')
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    cs = 0
    CFL0 = 1
    dim = 2
    nSections = 1
    c_ref = 1
    plotVars = ['H']

    ms = [.001, .01] # [LE, TE]

    # ========================================================================================

    # define flow variables
    Re, alpha, M_inf, arf = Case(cs)


    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()

    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1, 'msTe' : ms[1], 'msLe' : ms[0]}
    msh, writer = floD.mesh(dim, arf, pars, ['field', 'airfoil', 'downstream'])
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, blw = floD.problem(msh, dim, alpha, 0., M_inf, c_ref, c_ref, 0.25, 0., 0., 'airfoil', te = 'te', vsc = True, dbc=True)
    tms['pre'].stop()

    # solve viscous problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()

    # Initialize solvers and coupler
    
    iSolver = floD.newton(pbl)
    vSolver = viscU.initBL(Re, M_inf, CFL0, nSections)

    coupler = floVC.Coupler(iSolver, vSolver, blw)
    coupler.Run()

    # extract Cp
    Cp = floU.extract(bnd.groups[0].tag.elems, iSolver.Cp)
    # Extract BL variables
    vSln = viscU.GetSolution(0, vSolver)

    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('      Re        M    alpha       Cl       Cd      Cdp      Cdf       Cm')
    print('{0:6.1f}e6 {1:8.2f} {2:8.1f} {3:8.4f} {4:8.4f} {5:8.4f} {6:8.4f} {7:8.4f}\n'.format(Re/1e6, M_inf, alpha*180/math.pi, iSolver.Cl, vSolver.Cdt, vSolver.Cdp, vSolver.Cdf, iSolver.Cm))

    print(ccolors.ANSI_BLUE + 'PyWriting...' + ccolors.ANSI_RESET)
    tboxU.write(Cp, 'Cp_Corrected.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    viscU.WriteFile(vSln, vSolver.GetRe())
    print(' ')

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)
    print('SOLVERS statistics')
    print(coupler.tms)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if Re == 1e7 and M_inf == 0. and alpha == 2.*math.pi/180:
        tests.add(CTest('Cl', iSolver.Cl, 0.2208, 5e-3))
        tests.add(CTest('Cd', vSln['Cdt'], 0.00531, 1e-3, forceabs=True))
        tests.add(CTest('Cdp', vSln['Cdp'], 0.0009, 1e-3, forceabs=True))
        tests.add(CTest('xtr_top', vSln['xtrT'], 0.20, 0.03, forceabs=True))
        tests.add(CTest('xtr_bot', vSln['xtrB'], 0.50, 0.03, forceabs=True))
    if Re == 1e7 and M_inf == 0. and alpha == 5.*math.pi/180:
        tests.add(CTest('Cl', iSolver.Cl, 0.553, 5e-3))
        tests.add(CTest('Cd', vSln['Cdt'], 0.0062, 1e-3, forceabs=True))
        tests.add(CTest('Cdp', vSln['Cdp'],0.0018, 1e-3, forceabs=True))
        tests.add(CTest('xtr_top', vSln['xtrT'], 0.06, 0.03, forceabs=True))
        tests.add(CTest('xtr_bot', vSln['xtrB'], 0.74, 0.03, forceabs=True))
    else:
        print(ccolors.ANSI_RED + 'Test not defined for this flow' + ccolors.ANSI_RESET)

    tests.run()

    viscU.PlotVars(vSln, plotVars, cs)

    # eof
    print('')

if __name__ == "__main__":
    main()
